from typing import Callable, TypeVar, List

import functools
import re

from adhesive.steps.AdhesiveTask import AdhesiveTask
from adhesive.model.AdhesiveProcess import AdhesiveProcess
from adhesive.model.WorkflowExecutor import WorkflowExecutor

from adhesive.xml.bpmn import read_bpmn_file

T = TypeVar('T')
process = AdhesiveProcess()


def task(namere: str) -> Callable[..., Callable[..., T]]:
    def wrapper_builder(f: Callable[..., T]) -> Callable[..., T]:
        process.steps.append(AdhesiveTask(namere , f))
        return f

    return wrapper_builder


def bpmn_build(file_name: str):
    """ Start a build that was described in BPMN """
    process.workflow = read_bpmn_file(file_name)
    WorkflowExecutor(process).execute()




from typing import Dict, Iterable, List

from .Task import Task
from .StartEvent import StartEvent
from .EndEvent import EndEvent
from .Edge import Edge


class Workflow:
    """
    A workflow for the build
    """
    def __init__(self) -> None:
        self._start_events: Dict[str, StartEvent] = dict()
        self._tasks: Dict[str, Task] = dict()
        self._edges: Dict[str, Edge] = dict()
        self._end_events: Dict[str, EndEvent] = dict()

    @property
    def start_events(self) -> Dict[str, StartEvent]:
        return self._start_events

    @property
    def tasks(self) -> Dict[str, Task]:
        return self._tasks

    @property
    def edges(self) -> Dict[str, Edge]:
        return self._edges

    @property
    def end_events(self) -> Dict[str, EndEvent]:
        return self._end_events

    def add_task(self, task: Task) -> None:
        """ Add a task into the graph. """
        self._tasks[task.id] = task

    def add_edge(self, edge: Edge) -> None:
        """ Add an edge into the graph. """
        self._edges[edge._id] = edge

    def add_start_event(self, event: StartEvent) -> None:
        self._start_events[event._id] = event

    def add_end_event(self, event: EndEvent) -> None:
        self._end_events[event._id] = event

    def get_outgoing_edges(self, task_id: str) -> Iterable[Edge]:
        """ Get the outgoing edges. """
        result: List[Edge] = []

        for edge_id, edge in self._edges.items():
            if edge.source_id == task_id:
                result.append(edge)

        return result



from typing import Set, Optional, Dict, List

from adhesive.graph.Task import Task
from adhesive.steps.AdhesiveTask import AdhesiveTask
from adhesive.steps.WorkflowContext import WorkflowContext

from .AdhesiveProcess import AdhesiveProcess


class WorkflowExecutor:
    """
    An executorof AdhesiveProcesses.
    """
    def __init__(self,
                 process: AdhesiveProcess) -> None:
        self.process = process

    def execute(self) -> None:
        tasks_impl: Dict[str, AdhesiveTask] = dict()
        self._validate_tasks(tasks_impl)

        workflow = self.process.workflow
        active_events: List[str] = [ _id for _id in workflow.start_events.keys() ]

        while active_events:
            event_id = active_events.pop()
            self.process_event(tasks_impl, event_id)

            for outgoing_edge in workflow.get_outgoing_edges(event_id):
                active_events.append(outgoing_edge.target_id)

    def process_event(self, tasks_impl: Dict[str, AdhesiveTask], task_id: str) -> None:
        if task_id not in tasks_impl:
            return

        task = self.process.workflow.tasks[task_id]
        context = WorkflowContext(task)
        params = [context]

        tasks_impl[task_id].invoke(context)


    def _validate_tasks(self, tasks_impl) -> None:
        unmatched_tasks: Set[Task] = set()

        for task_id, task in self.process.workflow.tasks.items():
            adhesive_step = self._match_task(task)

            tasks_impl[task_id] = adhesive_step

            if not adhesive_step:
                unmatched_tasks.add(task)

        if unmatched_tasks:
            print("Missing tasks implementations. Generate with:\n")
            for unmatched_task in unmatched_tasks:
                print(f"@adhesive.task('{unmatched_task.name}')")
                print("def task_impl(context):")
                print("    pass\n\n")

            raise Exception("Missing tasks implementations")

    def _match_task(self, task: Task) -> Optional[AdhesiveTask]:
        for step in self.process.steps:
            if step.matches(task.name) is not None:
                return step

        return None





class StartEvent:
    """
    StartEvent documentation.
    """
    def __init__(self,
                 _id: str,
                 name: str) -> None:
        self._id = _id
        self.name = name




from typing import Callable, Optional, List
import re

from .WorkflowContext import WorkflowContext


class AdhesiveTask:
    """
    A task implementation.
    """
    def __init__(self,
                 expression: str,
                 code: Callable) -> None:
        self.re_expression = re.compile(expression)
        self.code = code
        pass

    def matches(self, name: str) -> Optional[List[str]]:
        m = self.re_expression.match(name)

        if not m:
            return None

        return list(m.groups())

    def invoke(self,
        context: WorkflowContext) -> None:
        step_name = context.task.name
        params = self.matches(step_name)

        self.code(context, *params)  # type: ignore




import unittest
from adhesive.xml.bpmn import read_bpmn_file
from adhesive.graph.Workflow import Workflow


class TestReadingBpmn(unittest.TestCase):
    """
    Test if we can read a BPMN file correctly.
    """

    def test_reading_bpmn(self) -> None:
        """
        Try to see if reading a basic BPMN works.
        """
        workflow = read_bpmn_file("test/adhesive/xml/adhesive.bpmn")

        self.assertEqual(4, len(workflow.tasks))
        self.assertEqual(6, len(workflow.edges))
        self.assertEqual(1, len(workflow.start_events))
        self.assertEqual(1, len(workflow.end_events))

        first_task = workflow.tasks["_3"]
        self.assertEqual("Build Germanium Image", first_task.name)

        self.assertTrue(workflow)


if __name__ == '__main__':
    unittest.main()



import unittest
import adhesive

from adhesive.model.AdhesiveProcess import AdhesiveProcess
from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.xml.bpmn import read_bpmn_file


class TestWorkflowExecutor(unittest.TestCase):
    """
    Test if the workflow executor can execute simple workflows.
    """

    def test_simple_execution(self):
        """
        Load a simple workflow and execute it.
        """
        adhesive.process = AdhesiveProcess()
        executed_steps: List[str] = []

        @adhesive.task('.*')
        def task_impl(context) -> None:
            executed_steps.append(context.task.name)

        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/adhesive.bpmn")

        WorkflowExecutor(adhesive.process).execute()
        self.assertEqual(executed_steps, [
            "Ensure Docker Tooling",
            "Build Germanium Image",
            "Test Firefox",
            "Test Chrome",
        ])


if __name__ == '__main__':
    unittest.main()



from typing import Tuple

from xml.etree import ElementTree
import re

from adhesive.graph.Workflow import Workflow
from adhesive.graph.Task import Task
from adhesive.graph.Edge import Edge
from adhesive.graph.StartEvent import StartEvent
from adhesive.graph.EndEvent import EndEvent

TAG_NAME = re.compile(r'^(\{.+\})?(.+)$')
SPACE = re.compile(r"\s+", re.IGNORECASE)


def read_bpmn_file(file_name: str) -> Workflow:
    """ Read a BPMN file as a build workflow. """
    result = Workflow()

    root_node = ElementTree.parse(file_name).getroot()
    process = find_node(root_node, 'process')

    for node in process.getchildren():
        process_node(result, node)

    return result


def find_node(parent_node, name: str):
    for node in parent_node.getchildren():
        _, node_name = parse_tag(name)
        if node_name == name:
            return node

    return None


def process_node(result: Workflow,
                 node) -> None:
    node_ns, node_name = parse_tag(node.tag)

    if "task" == node_name:
        process_node_task(result, node)
    elif "sequenceFlow" == node_name:
        process_node_sequence_flow(result, node)
    elif "startEvent" == node_name:
        process_node_start_event(result, node)
    elif "endEvent" == node_name:
        process_node_end_event(result, node)
    else:
        print(f"{node_name} node ignored")


def process_node_task(w: Workflow, xml_node) -> None:
    """ Create a Task element from the workflow """
    node_name = normalize_name(xml_node.get("name"))
    task = Task(xml_node.get("id"), node_name)
    w.add_task(task)


def process_node_start_event(w: Workflow, xml_node) -> None:
    """ Create a start event from the workflow """
    node_name = normalize_name(xml_node.get("name"))
    task = StartEvent(xml_node.get("id"), node_name)
    w.add_start_event(task)


def process_node_end_event(w: Workflow, xml_node) -> None:
    """ Create an end event from the workflow """
    node_name = normalize_name(xml_node.get("name"))
    task = EndEvent(xml_node.get("id"), node_name)
    w.add_end_event(task)


def process_node_sequence_flow(w: Workflow, xml_node) -> None:
    edge = Edge(xml_node.get("id"),
                xml_node.get("sourceRef"),
                xml_node.get("targetRef"))
    w.add_edge(edge)


def normalize_name(name: str) -> str:
    return SPACE.sub(' ', name)


def parse_tag(name: str) -> Tuple[str, str]:
    m = TAG_NAME.match(name)

    if not m:
        raise Exception(f"Unable to parse tag name `{name}`")

    return m[1], m[2]

